#ifndef GRID_HEALPIX_H
#define GRID_HEALPIX_H

#include <cinttypes>

enum class HpOrder
  {
    Undef,
    XY,
    Ring,
    Nest
  };

HpOrder hp_get_order(const char *orderName);
void hp_generate_coords(HpOrder order, int nside, int64_t nvals, double *xvals, double *yvals, bool withBounds, double *xbounds, double *ybounds);

#endif
